﻿#nowarn "864"

namespace Chapter2RealWorldSample

open System.Collections.Generic
open System.Collections.Specialized

// define ObservableList<'T> class
type ObservableList<'T> () = 
    inherit System.Collections.Generic.List<'T>()     //inherit List
        
    let collectionChanged = Event< _ , _ > ( )

    let mutable isTriggerEvent = false   //mutable variable

    // additional constructor to initialize the instance with given items
    new(items) = ObservableList<'T>() then base.AddRange(items)

    // implement the INotifyCollectionChanged interface
    interface INotifyCollectionChanged with
        [<  CLIEvent  >]
        member this.CollectionChanged = collectionChanged.Publish

    // property to control if any of the events should be triggered
    member this.IsTriggerEvent
        with get() = isTriggerEvent
        and set(v) = isTriggerEvent <- v

    //add element
    member this.Add(n:'T) = 
        base.Add(n)
        this.TriggerAdd(n, base.Count)

    // remove element
    member this.Remove(n:'T) = 
        let index = this.IndexOf(n)
        if index <> -1 then
            let r = base.Remove(n)
            this.TriggerRemove(n, index)
            r
        else 
            false

    member this.Remove(n:'T seq) =
        n |> Seq.iter (this.Remove >> ignore)
    member this.RemoveAt(index) = 
        if this.Count > 0 then 
            if index >=0 && index < this.Count then
                let item = base.[index]
                base.RemoveAt(index)
                this.TriggerRemove(item, index)
    member this.RemoveRange(index, count) = 
        [0..count-1] |> Seq.iter (fun i -> this.RemoveAt(index))

    // sort elements
    member this.Sort()= 
        base.Sort()
        this.TriggerReset()
    member this.Sort(comparer:IComparer<'T>) = 
        base.Sort(comparer)
        this.TriggerReset()
    member this.Sort(index, count, comparer) = 
        base.Sort(index, count, comparer)
        this.TriggerReset()

    // reverse elements in the list
    member this.Reverse() = 
        base.Reverse()
        this.TriggerReset()
    member this.Reverse(index, count) = 
        base.Reverse(index, count)
        this.TriggerReset()

    // add items
    member this.AddRange(items) = 
        items |> Seq.iter this.Add

    // clear current list
    member this.Clear() = 
        base.Clear()
        this.TriggerReset()

    // insert element at the index location
    member this.Insert(index, item) = 
        base.Insert(index, item)
        this.TriggerAdd(item, index)

    // insert multiple elements at the index location
    member this.InsertRange(index, items) = 
       items |> Seq.iteri (fun i item -> this.Insert(index+i, item))

    // define an indexer
    member this.Item
        with get(i) = base.[i]
        and set i v = 
            let old = base.[i]
            base.[i] <- v
            this.TriggerReplace(v, old, i)

    // trigger add, remove, replace, and reset events
    member private this.TriggerAdd(item, index) =
        if this.IsTriggerEvent then
            collectionChanged.Trigger(this, 
                NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, item, index))
    member private this.TriggerRemove(item, index) =
        if this.IsTriggerEvent then
            collectionChanged.Trigger(this, 
                NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, item, index))
    member private this.TriggerReplace(item, oldItem, index) =
        if this.IsTriggerEvent then
            collectionChanged.Trigger(this,    NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Replace, item, oldItem, index))
    member public this.TriggerReset() =
        if this.IsTriggerEvent then
            collectionChanged.Trigger(this, 
                NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset, null, -1))

    // static function to convert any IEnumerable variable
    static member FromList(items) = ObservableList(items)


